module QueryInterface
  module Server
    class LazyQuery

      attr_accessor :model, :result_model, :transformations

      def initialize(model, transformations=nil, result_model=nil)
        self.model = model
        if transformations
          self.transformations = transformations.map {|item| item.dup}
        else
          self.transformations = []
        end
        self.result_model = result_model
      end

      def parse(data)
        (self.result_model ? self.result_model.parse(data) : self.model.parse(data))
      end

      def instantiate(data)
        (self.result_model ? self.result_model.new(data) : self.model.new(data))
      end

      def instantiate_collection(parsed_data)
        (self.result_model ? self.result_model.new_collection(parsed_data) : self.model.new_collection(parsed_data))
      end

      def copy(options = {})
        self.class.new(self.model, self.transformations, self.result_model)
      end

      def add_transformation(type, parameter=nil)
        self.transformations << {'transformation' => type, 'parameter' => parameter}
      end

      def exclude(conditions={})
        self.copy.tap do |query|
          conditions.each do |key, value|
            query.add_transformation("exclude", {"field" => key, "value" => value})
          end
        end
      end

      def filter(conditions={})
        self.copy.tap do |query|
          conditions.each do |key, value|
            query.add_transformation("filter", {"field" => key, "value" => value})
          end
        end
      end

      def instance(id)
        self.copy.tap do |query|
          query.add_transformation("instance", id)
        end
      end

      def context(association, model=nil)
        self.copy.tap do |query|
          query.result_model = (model ? model : association.to_s.singularize.camelize.constantize)
          query.add_transformation("context", association.to_s)
        end
      end

      def with(*fields, **opts)
        self.copy.tap do |query|
          fields.each do |field|
            query.add_transformation("with", {"field" => field.to_s})
          end
          opts.each do |field, param|
            query.add_transformation("with", {"field" => field.to_s, "param" => param})
          end
        end
      end

      def order(*fields)
        self.copy.tap do |query|
          fields.each do |field|
            query.add_transformation("order", field)
          end
        end
      end

      def evaluate
        self.do_query
      end

      def dataset
        self.do_query(false)
      end

      def ids
        query = self.copy
        query.add_transformation("map_ids")
        query.do_query
      end

      def count
        query = self.copy
        query.add_transformation("count")
        query.do_query[:count]
      end

      def update(data)
        query = self.copy
        query.add_transformation("update", data)
        query.do_query[:updated_count]
      end

      def do_query(add_default_order=true)
        transformer = Transformations::SequelTransformer.new(self.model.filter)
        transformer.run(self.transformations, add_default_order)
      end

      def first(*args)
        one("first", *args)
      end

      def last(*args)
        one("last", *args)
      end

      def to_json(*args)
        evaluate.to_json(*args)
      end

      def method_missing(method_name, *args, &block)
        evaluate.send(method_name, *args, &block)
      end
    protected

      def one(which, params = {})
        query = self.copy
        query.add_transformation(which)
        query.do_query
      end
    end
  end
end
